//  Copyright (c) 2003, Sarwat Khan. All rights reserved.
//  
//  Redistribution and use in source and binary forms, with or without
//  modification, are permitted provided that the following conditions are
//  met:
//          
//      1. Redistributions of source code must retain the above copyright
//         notice, this list of conditions and the following disclaimer.
//          
//      2. Redistributions in binary form must reproduce the above
//         copyright notice, this list of conditions and the following disclaimer
//         in the documentation and/or other materials provided with the
//         distribution.
//          
//      3. Neither Sarwat Khans name nor the names of any contributors to
//         the Tomato Torrent Software may be used to endorse or promote
//         products derived from this software without specific prior written
//         permission.
//  
//  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
//  "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
//  LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
//  A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE COPYRIGHT
//  OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
//  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
//  LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
//  DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
//  THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
//  (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
//  OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.


#import "StartingPointController.h"

#import "DocumentController.h"
#import "TrackerWindowController.h"
#import "NSFileManager+LocalizedPaths.h"

@interface StartingPointController (Private)
- _hideWindow;
- _showWindow;
- _updateRecentTorrentsMenu;
- (void) _checkNumWindows;
- (void) _windowWillClose:notif;
- (void) _windowDidBecomeMain:notif;
@end

@implementation StartingPointController

+ instance
{
	static id instance = nil;
	
	if (instance == nil)
		instance = [self new];
	
	return instance;
}

- init
{
	[[NSNotificationCenter defaultCenter] addObserver:self
											 selector:@selector(_windowWillClose:)
												 name:NSWindowWillCloseNotification
											   object:nil];

	[[NSNotificationCenter defaultCenter] addObserver:self
											 selector:@selector(_windowDidBecomeMain:)
												 name:NSWindowDidBecomeMainNotification
											   object:nil];
											
	return [super initWithWindowNibName:@"StartingPoint"];
}

- (void) dealloc
{
	[[NSNotificationCenter defaultCenter] removeObserver:self];
	
	[super dealloc];
}

- showStartingPointWindow
{
	[self _showWindow];
	return self;
}

- (IBAction)notImplemented:(id)sender
{
	NSLog(@"not implemented");
}

- (IBAction) showTrackerStartingPoint:sender
{
	[[TrackerWindowController new] showWindow:nil];
}

- (IBAction)openRecentTorrent:(id)sender
{
	id	torrent;
	
	if ([_recentTorrentsMenu numberOfItems] > 0)
	{
		torrent = [[_recentTorrentsMenu selectedItem] representedObject];
		
		[[NSDocumentController sharedDocumentController] 
			openDocumentWithContentsOfURL:torrent 
								  display:YES];
	}
}

- (IBAction)recentTorrentSelected:(id)sender
{
	NSRect		frame, windowFrame;
	float		windowMaxX, frameMaxX;
	
	//adjust the thing to fit.
	[_recentTorrentsMenu sizeToFit];
	
	frame = [_recentTorrentsMenu convertRect:[_recentTorrentsMenu bounds] toView:[[self window] contentView]];
	windowFrame = [[[self window] contentView] frame];

	windowMaxX = NSMaxX(windowFrame) - 20;
	frameMaxX = NSMaxX(frame);
	
	if (frameMaxX > windowMaxX)
	{
		frame.size.width -= frameMaxX - windowMaxX;
		[_recentTorrentsMenu setFrameSize:frame.size];
	}
	
	[[_recentTorrentsMenu superview] setNeedsDisplay:YES];
}

@end

@implementation StartingPointController (Private)

- (void) _windowWillClose:notif
{
	if ([notif object] != [self window])
		if ([NSApp isActive])
			[self performSelector:@selector(_checkNumWindows) withObject:nil afterDelay:0.5];
}

- (void) _windowDidBecomeMain:notif
{
	if ([notif object] != [self window])
		if ([[notif object] isExcludedFromWindowsMenu] == NO)
			[self _hideWindow];
}

- (void) _checkNumWindows;
{
	id		window, ite;
	int		numWindowsVisible = 0;
	//get count of visible windows
	ite = [[NSApp windows] objectEnumerator];
	while (window = [ite nextObject])
		if ([window isVisible])	numWindowsVisible++;
	
	if (numWindowsVisible == 0)
		[self _showWindow];
}

- _hideWindow
{
	[[self window] orderOut:nil];
	return self;
}

- _showWindow
{
	[self window];	//load.
	
	[self _updateRecentTorrentsMenu];
	[[self window] center];
	[[self window] makeKeyAndOrderFront:nil];
	return self;
}

static id _updateRecentTorrentsMenu_trim(id path)
{
	return [path stringByDeletingPathExtension];
}

- _updateRecentTorrentsMenu
{
	id		title, titles, ite, url, menuItem, menuItems, urls, menu;
	id		docController = [NSDocumentController sharedDocumentController];
	id		fileManager = [NSFileManager defaultManager];
			
	[_recentTorrentsMenu removeAllItems];
	urls = [docController recentDocumentURLs];
	
	[_recentTorrentsMenu setEnabled:NO];

	if ([urls count] > 0)
	{
		menuItems = [NSMutableArray array];
		titles = [NSCountedSet set];
		
		//pass one: make menu items for every URL.
		ite = [urls objectEnumerator];
		while (url = [ite nextObject])
		{
			if ([url isFileURL])
			{
				title = nil;
				
				if ([fileManager fileExistsAtPath:[url path]])
					title = _updateRecentTorrentsMenu_trim([[url path] lastPathComponent]);
			}
			else					
			{
				title = [NSString stringWithFormat:
							SKLocalizedString(
								@"%@ @ %@", 
								@"(lastPathComponent) at (url host), recent torrent url"),
							_updateRecentTorrentsMenu_trim([[url path] lastPathComponent]),
							[url host]];
			}
			
			if (title)
			{
				menuItem = [[NSMenuItem new] autorelease];
				[menuItem setRepresentedObject:url];
				[menuItem setTitle:title];
	
				[menuItems addObject:menuItem];
				[titles addObject:title];
			}
		}

		if ([menuItems count] > 0)
		{
			//pass two: handle duplicate titles.
			ite = [menuItems objectEnumerator];
			while (menuItem = [ite nextObject])
			{
				title = [menuItem title];
				if ([titles countForObject:title] > 1)
				{
					url = [menuItem representedObject];
					if ([url isFileURL])
					{
						title = [NSString stringWithFormat:
							SKLocalizedString(@"%@ - %@", @"(file name) mdash (short path to it) : duplicate file name (recent urls)"),
							title,
							_updateRecentTorrentsMenu_trim([fileManager displayPathForPath:[url path]])];	//AppKit's recent menu is much smarter than this.
					}
					else
					{
						//gurg. this is rare, I'd think. I'm just going to put out the whole url minus lastPathComponent.
						title = [NSString stringWithFormat:
							SKLocalizedString(@"%@ - %@",  @"(url file name) mdash (path to file) : duplicate url (recent url menu item)"),
							_updateRecentTorrentsMenu_trim([[url path] lastPathComponent]),
							[[url absoluteString] stringByDeletingLastPathComponent]];	//this screws up the url a bit
																						//as it tries to return a valid file path.
					}
					
					[menuItem setTitle:title];
				}
			}
			//now add the menu items to the popup.
			menu = [_recentTorrentsMenu menu];
	
			ite = [menuItems objectEnumerator];		
			while (menuItem = [ite nextObject]) [menu addItem:menuItem];
	
			[_recentTorrentsMenu selectItemAtIndex:0];
			[self recentTorrentSelected:nil];
			[_recentTorrentsMenu setEnabled:YES];
		}
	}
	
	return self;
}

@end

