//  Copyright (c) 2003, Sarwat Khan. All rights reserved.
//  
//  Redistribution and use in source and binary forms, with or without
//  modification, are permitted provided that the following conditions are
//  met:
//          
//      1. Redistributions of source code must retain the above copyright
//         notice, this list of conditions and the following disclaimer.
//          
//      2. Redistributions in binary form must reproduce the above
//         copyright notice, this list of conditions and the following disclaimer
//         in the documentation and/or other materials provided with the
//         distribution.
//          
//      3. Neither Sarwat Khans name nor the names of any contributors to
//         the Tomato Torrent Software may be used to endorse or promote
//         products derived from this software without specific prior written
//         permission.
//  
//  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
//  "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
//  LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
//  A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE COPYRIGHT
//  OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
//  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
//  LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
//  DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
//  THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
//  (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
//  OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.


#import "TrackerWindowController.h"

#import "TrackerDocument.h"
#import "TrackerDocumentController.h"

#import "NSWindow+WindowContentSize.h"

@interface TrackerWindowController (Private)
- (void) _openDocument:url;
- _installStartingPointView;
@end

@implementation TrackerWindowController
- init
{
	self = [super initWithWindowNibName:@"Tracker"];

	return self;
}

- (void) dealloc
{
	[_saveEmptyView release];
	
	[super dealloc];
}

- (void) windowDidLoad
{
	_saveEmptyView = [[[self window] contentView] retain];
	
	if ([self document])
		[_trackerDocumentController setupWindowForDocumentState];
	else
		[self _installStartingPointView];
}

- (void) windowWillClose:notif
{
	if ([self document] == nil)
		[self autorelease];
}

- trackerDocumentController
{
	[self window];	//load nib.
	return _trackerDocumentController;
}

- (void)synchronizeWindowTitleWithDocumentName;
{
	[super synchronizeWindowTitleWithDocumentName];

	if ([self document] != nil)	//I would expect that's redundant.
	{
		[_trackerDocumentController documentIsSpecified:[[self document] fileName] != nil];
	}
}

- (void)showWindow:sender
{
	if ([self document] == nil)
	{
		[[self window] center];
		[[self window] makeKeyAndOrderFront:nil];
	}
	else
	{
		[self setShouldCascadeWindows:YES];
		[[self window] makeKeyAndOrderFront:nil];
	}
}

- showView:view;
{
	NSRect		frame;
	BOOL		isVisible = [[self window] isVisible];
	
	//move the current view out before sizing
	[[self window] setContentView:_saveEmptyView];
	
	frame = [view frame];
	//[[self window] setContentSize:frame.size];		//I'd like to do that with animation.
	[[self window] setContentSize:frame.size display:isVisible animate:isVisible];

	//set the view now that the window's the right size.
	[[self window] setContentView:view];
	
	return self;
}

- (IBAction) openUntitledTrackerDocument:sender;
{
	[self _openDocument:nil];
}

- (IBAction) openTrackerDocument:sender;
{
	id		panel = [NSOpenPanel openPanel];
	
	[panel setAllowsMultipleSelection:NO];
	[panel beginSheetForDirectory:NSHomeDirectory() 
							 file:nil
							types:[NSArray arrayWithObject:kTrackerDocumentExtension]
				   modalForWindow:[self window]
				    modalDelegate:self 
				   didEndSelector:@selector(_trackerOpenPanelDidEnd:returnCode:contextInfo:) 
					  contextInfo:nil];
}

- (IBAction) openLastTrackerDocument:sender;
{
	id	url;
	
	if (url = [TrackerDocument lastTrackerDocumentURL])
	{
		[self _openDocument:url];
	}
}

- (void)_trackerOpenPanelDidEnd:(NSOpenPanel *)sheet returnCode:(int)returnCode contextInfo:(void  *)contextInfo
{
	if (returnCode == NSOKButton)
		[self _openDocument:[[sheet URLs] lastObject]];
}

@end

@implementation TrackerWindowController (Private)

- (void) _openDocument:url
{
	id		documentController, doc;
	BOOL	shouldCreate;
	
	documentController = [NSDocumentController sharedDocumentController];
	shouldCreate = [documentController shouldCreateUI];
	[documentController setShouldCreateUI:NO];
	
	if (url)
		doc = [documentController openDocumentWithContentsOfURL:url display:NO];
	else
		doc = [documentController openUntitledDocumentOfType:kTrackerDocumentType display:NO];

	if (doc)
	{
		if ([[doc windowControllers] count] == 0)
		{
			[doc setTrackerWindowController:self];
			[_trackerDocumentController setupWindowForDocumentState];
			[doc showWindows];
		}
		else
		{
			[[self window] orderOut:nil];
			[doc showWindows];
			[self close];
		}
	}
	
	[documentController setShouldCreateUI:shouldCreate];
}

- _installStartingPointView;
{
	id		view, lastURL, lastURLString;
	BOOL	enable;
	NSRect	frame;

	view = _startingPointView;
	frame = [view frame];
	
	//set content view and setup window
	[self showView:view];

	[[self window] setTitle:NSLocalizedString(@"Tracker Starting Point", @"tracker starting point window title")];
	[[self window] setMinimumContentSize:frame.size maximumSize:frame.size];
	[[self window] center];

	enable = NO;
	lastURLString = @"";
	if (lastURL = [TrackerDocument lastTrackerDocumentURL])
	{
		if ([lastURL isFileURL] && [[NSFileManager defaultManager] fileExistsAtPath:[lastURL path]])
		{
			enable = YES;
			lastURLString = [[[lastURL path] stringByAbbreviatingWithTildeInPath] stringByDeletingPathExtension];
		}
	}

	[_lastTrackerFilePath setStringValue:lastURLString];
	[_openLastTrackerButton setEnabled:enable];

	return self;	
}

- (BOOL)windowShouldZoom:(NSWindow *)window toFrame:(NSRect)newFrame;
{
	return [window contentView] != _startingPointView;
}

@end
