//==============================================================================
//		ttagent_lib.c											TomatoTorrent
//==============================================================================
//  Created by Sarwat Khan on Wed Jun 04 2003.
//  Copyright (c) 2003, Sarwat Khan. All rights reserved.
//  
//  Redistribution and use in source and binary forms, with or without
//  modification, are permitted provided that the following conditions are
//  met:
//          
//      1. Redistributions of source code must retain the above copyright
//         notice, this list of conditions and the following disclaimer.
//          
//      2. Redistributions in binary form must reproduce the above
//         copyright notice, this list of conditions and the following disclaimer
//         in the documentation and/or other materials provided with the
//         distribution.
//          
//      3. Neither Sarwat Khan's name nor the names of any contributors to
//         the Tomato Torrent Software may be used to endorse or promote
//         products derived from this software without specific prior written
//         permission.
//  
//  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
//  "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
//  LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
//  A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE COPYRIGHT
//  OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
//  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
//  LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
//  DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
//  THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
//  (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
//  OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
//
//	Code that is pretty much posix. And then some BSD.

#include "tomato_agent.h"
#include "tomato_agent_logging.h"

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <pthread.h>
#include <fcntl.h>

#include <sys/syslog.h>
#include <sys/time.h>

//------------------------------------------------------------------------------
//		 constants
//------------------------------------------------------------------------------


//code assumes that last character (#11) is insignificant
#define k_ttagent_num_answer_length_bytes		11	
#define k_ttagent_shutdown_message				"shutdown\n"

#define k_ttagent_pyscript_agentboot			"import TomatoTorrentPosixAgent; TomatoTorrentPosixAgent.main()"

//------------------------------------------------------------------------------
//		 structs
//------------------------------------------------------------------------------

struct ttagent_posix_context
{
	pthread_t			stderr_thread;
	pthread_mutex_t		tail_mutex;
	pthread_cond_t		tail_signal;
	bool				tail_halt;
};
typedef struct ttagent_posix_context ttagent_posix_context;

// this is one of those things where, when you come back to the code
// after a few years to review it, you say, WTF did I need to do this?
struct ttagent_priv_startup_fmts
{
	char	*noslash;
	char	*slash;
};

#define k_ttagent_python_libroot		0		//for indexing into the thing below

static const struct ttagent_priv_startup_fmts k_ttagent_startup_pylibdir_pypaths[] = 
{
	{"%sTomatoTorrentAgent/%s", "%s/TomatoTorrentAgent/%s"}, 
	{"%sTomatoTorrentAgent/%s/" k_ttagent_python_zipmodule, "%s/TomatoTorrentAgent/%s/" k_ttagent_python_zipmodule}
};
	
//------------------------------------------------------------------------------
//		 private functions
//------------------------------------------------------------------------------

static void		ttagent_priv_set_error(ttagent_context *agent, char *error);
static size_t	ttagent_priv_read(int input, char* buffer, unsigned length);

static void 	ttagent_priv_posix_startup_errortail(ttagent_context *agent);
static void 	ttagent_priv_posix_shutdown_errortail(ttagent_context *agent);
static void*	ttagent_priv_posix_tail_stderr(void *agent);


//------------------------------------------------------------------------------
//		 ttagent_startup
//------------------------------------------------------------------------------
//	spawns the new process of exec_name in directory exec_dir. exec_name is likely
//	TomatoTorrentAgent.py.
//
//	Pass NULL for out_error_desc if you don't care about error descriptions. 
//	If you (*out_error_desc) is not NULL on return, you should free the string.
//

ttagent_context*	
ttagent_startup(const char *resource_dir, char** out_error_desc)
{
	ttagent_context		*agent;
	int					ttagent_read_fds[2],
						ttagent_write_fds[2],
						ttagent_error_fds[2];
	char				*error_text;
	char				*format;
	char				*pylib_path;
	char				*exec_dir;
	pid_t				pid;
	pid_t				parent_pid;
	bool				fudge = false;
	int					result;

	error_text = NULL;
	if (resource_dir[strlen(resource_dir)-1] != '/') fudge = true; 
	
	agent = (ttagent_context*)malloc(sizeof(ttagent_context));
	agent->pid = 0;
		
	//define the directory that contains the bittorrent modules
	//
	
	//use pylib_path to define PYTHONPATH
	format 				= k_ttagent_startup_pylibdir_pypaths[g_ttagent_python_zipimport].noslash;
	if (fudge) format 	= k_ttagent_startup_pylibdir_pypaths[g_ttagent_python_zipimport].slash;
	
	asprintf(&pylib_path, format, resource_dir, g_ttagent_python_sourcedir);
		setenv("PYTHONPATH", pylib_path, 1);
		ttagent_logging_debug(
			-1,//kTomatoAgentLoggingLevel_Informative, 
			"ttagent_startup: PYTHONPATH=%s", 
			pylib_path);
	free(pylib_path);

	//define exec_dir, which will be the cwd.
	//
	
	format 				= k_ttagent_startup_pylibdir_pypaths[k_ttagent_python_libroot].noslash;
	if (fudge) format 	= k_ttagent_startup_pylibdir_pypaths[k_ttagent_python_libroot].slash;
	
	asprintf(&exec_dir, format, resource_dir, g_ttagent_python_sourcedir);
	
	//create file descriptors to communicate with
	//
	
	result = pipe(ttagent_read_fds);
	if (result == 0) result = pipe(ttagent_write_fds);
	if (result == 0) result = pipe(ttagent_error_fds);	
	if (result == 0)
	{
		//and fork to spawn the new process.
		parent_pid = getpid();
		pid = vfork();
		
		if (pid > 0)
		{
			//this code is executing in the original process, and pid is the 
			//pid of the child process which has already exec'd (with fork(),
			//these two would be running at the same time).
			
			agent->pid = pid;
			agent->read_fd = ttagent_read_fds[0];
			agent->write_fd = ttagent_write_fds[1];
			agent->error_fd = ttagent_error_fds[0];
			agent->last_error_desc = NULL;
			
			//close unused ends of pipes
			/*close(ttagent_read_fds[1]);
			close(ttagent_write_fds[0]);
			close(ttagent_error_fds[1]);*/
		}
		else if (pid == 0)
		{
			//this code is executing in the child process but using the parent's 
			//address space and stuff; parent is suspended. execl will change that.
			
			//set the process group so it can be killed with all the other agents
			setpgid(0, parent_pid);
			
			//reassign stdin, stdout, stderr to the fds from pipe()
			result = dup2(ttagent_write_fds[0], STDIN_FILENO);
			if (result >= 0) result = dup2(ttagent_read_fds[1], STDOUT_FILENO);
			if (result >= 0) result = dup2(ttagent_error_fds[1], STDERR_FILENO);
			
			//close unused ends of the pipes
			/*close(ttagent_write_fds[1]);
			close(ttagent_read_fds[0]);
			close(ttagent_error_fds[0]);*/
			
			if (result >= 0)
			{
				result = chdir(exec_dir);

				if (result == 0)
						syslog(LOG_INFO, "execing %s %s %s", k_ttagent_python_path, "-c", k_ttagent_pyscript_agentboot),
						ttagent_logging_debug(kTomatoAgentLoggingLevel_Motion, 
							"execing %s %s %s", k_ttagent_python_path, "-c", k_ttagent_pyscript_agentboot),
						//fprintf(stderr, "execing %s %s %s", k_ttagent_python_path, "-c", k_ttagent_pyscript_agentboot),
						execl(k_ttagent_python_path, k_ttagent_python_path, "-c", k_ttagent_pyscript_agentboot, (char*)NULL);
					
				else
					fprintf(stderr, "couldn't chdir %s\n", exec_dir),
					syslog(LOG_ERR, "couldn't chdir %s", exec_dir);
				
				fprintf(stderr, "couldn't exec %s\n", k_ttagent_python_path),
				syslog(LOG_ERR, "couldn't exec %s", k_ttagent_python_path);
			}
			else
				fprintf(stderr, "couldn't reassign fds.\n"),
				syslog(LOG_ERR, "couldn't reassign fds.");
				
			_exit(1);
		}
		else
		{
			error_text = "ttagent_startup: couldn't vfork child";
			fprintf(stderr, "%s\n", error_text);
		}
	}
	else
		error_text = "ttagent_startup: couldn't create file descriptors with pipe.";
		
	if (error_text)
	{
		free(agent);
		agent = NULL;
		
		if (out_error_desc)
			*out_error_desc = strdup(error_text);
	}
	
	free(exec_dir);
	
	if (agent)
		ttagent_priv_posix_startup_errortail(agent);
		
	return agent;
}

//------------------------------------------------------------------------------
//		 ttagent_shutdown
//------------------------------------------------------------------------------
//	sends the shutdown command to the process and closes the file descriptors.

void
ttagent_shutdown(ttagent_context *agent)
{
	char*		shutdown;
	
	shutdown = k_ttagent_shutdown_message;
	ttagent_send_message(agent, shutdown);
	kill(agent->pid, SIGHUP);
	
	ttagent_priv_posix_shutdown_errortail(agent);
	
	ttagent_priv_set_error(agent, NULL);
	free(agent);
}

//------------------------------------------------------------------------------
//		 ttagent_send_message
//------------------------------------------------------------------------------
//	see ttagent_send_message_fd

bool
ttagent_send_message(ttagent_context *agent, const char *message)
{
	char*		error_desc = NULL;
	bool		result;
	
	result = ttagent_send_message_fd(agent->write_fd, message, &error_desc);
	
	if (error_desc)
		ttagent_priv_set_error(agent, error_desc);
		
	return result;
}

//------------------------------------------------------------------------------
//		 ttagent_read_answer
//------------------------------------------------------------------------------
//	see ttagent_read_answer_fd

char*
ttagent_read_answer(ttagent_context *agent, char** answer)
{
	char*		error_desc = NULL;
	char*		result;
	
	result = ttagent_read_answer_fd(agent->read_fd, answer, &error_desc);
	
	if (error_desc)
		ttagent_priv_set_error(agent, error_desc);
		
	return result;
}

//------------------------------------------------------------------------------
//		 ttagent_send_message_fd
//------------------------------------------------------------------------------
//	sends the message to the agent. message doesn't have to be terminated with
//	a newline. If it returns false, you're probably hosed. 
//
//	Supply NULL for error desc if you don't want it. If *error_desc is non-NULL
//	on exit, you should free it.

bool
ttagent_send_message_fd(int output, const char *in_message, char **error_desc)
{
	unsigned		message_len;
	char*			message;
	bool			dispose;
	size_t			result;
	
	dispose = false;
	message = (char*)in_message;
	message_len = strlen(message);
	
	//add newline if needed.
	if (message[message_len-1] != '\n')
	{
		message = (char*)malloc(message_len+2);
		strcpy(message, in_message);
		message[message_len] = '\n';
		message[message_len+1] = 0;
		++message_len;
		dispose = true;
	}

	result = write(output, message, message_len);
	if (result != message_len && error_desc != NULL)
	{
		char	*error_format = "ttagent_send_message_fd: 'write' returned %d "
								"when trying to write %d bytes for message:\n%s";
		char	error_temp[255];
		
		snprintf(error_temp, 255, error_format, result, message_len, message);
		
		(*error_desc) = strdup(error_temp);
	}
	
	if (dispose) free(message);
	
	return result == message_len;
}

//------------------------------------------------------------------------------
//		 ttagent_read_answer_fd
//------------------------------------------------------------------------------
//	reads the message from the bugger. answer is stored in 'answer,' and it 
//	manages the memory for answer. That is, if *answer is NULL, it will allocate
//	memory for the result and return it, which must be freed. However, if *answer
//	is not NULL, it frees it and allocates a new one.
//
//	On error, returns nil. You're probably hosed in that case.
//
//	Supply NULL for error desc if you don't want it. If *error_desc is non-NULL
//	on exit, you should free it.

char*
ttagent_read_answer_fd(int input, char** io_answer, char** out_error_desc)
{
	char		answer_length[k_ttagent_num_answer_length_bytes];
	char		*answer;
	char		error_buffer[255];
	unsigned	length;
	size_t		result = 0;

	answer = NULL;
	if (io_answer)
	{
		if (*io_answer) 
		{
			free(*io_answer);
			*io_answer = NULL;
		}
	}
	
	//read the first 11 bytes, which is the length of the answer response.
	//result = read(input, answer_length, k_ttagent_num_answer_length_bytes);
	result = ttagent_priv_read(input, answer_length, k_ttagent_num_answer_length_bytes);
	
	if (result)
	{
		length = strtoul(answer_length, NULL, k_ttagent_num_answer_length_bytes-1);	//last char ignored
		if (length > 0)
		{
			//good, we now have to read 'length' bytes from input. That is the result.
			answer = (char*)calloc(length+1, 1);
			
			result = ttagent_priv_read(input, answer, length);
			if (result == 0)
			{
				free(answer); answer = NULL;
				strcpy(error_buffer, "ttagent_read_answer_fd: No data available");
			}
		}
		else
		{
			snprintf(error_buffer, 255, 
				"ttagent_read_answer_fd: Can't understand answer length as an unsigned long: %11s", 
				answer_length);
		}
	}
	else
		strcpy(error_buffer, "ttagent_read_answer_fd: No data available");

	if (answer == NULL && out_error_desc != NULL)
		*out_error_desc = strdup(error_buffer),
		syslog(LOG_ERR, error_buffer);
	
	if (io_answer) *io_answer = answer;
	
	return answer;
}

//--- private
#pragma mark == private ==

//------------------------------------------------------------------------------
//		 ttagent_priv_set_error
//------------------------------------------------------------------------------

static void 
ttagent_priv_set_error(ttagent_context *agent, char *error)
{
	if (agent->last_error_desc)
		free(agent->last_error_desc);

	agent->last_error_desc = error;
}

static size_t 
ttagent_priv_read(int input, char* buffer, unsigned length)
{
	size_t result = 0;
	
	do
		result += read(input, buffer + result, length - result);
	while
		(result < length && result != 0);
		
	return result;
}

//------------------------------------------------------------------------------
//		 ttagent_priv_posix_startup_errortail
//------------------------------------------------------------------------------

static void 
ttagent_priv_posix_startup_errortail(ttagent_context *agent)
{
	ttagent_posix_context	*posix_agent;
	
	//create a stderr tail thread for the agent
	if (posix_agent = (ttagent_posix_context*)calloc(sizeof(ttagent_posix_context), 1))
	{
		agent->_extra = posix_agent;
		pthread_mutex_init(&posix_agent->tail_mutex, NULL);
		pthread_cond_init(&posix_agent->tail_signal, NULL);

		pthread_mutex_lock(&posix_agent->tail_mutex);		//lock so thread can't start yet.
			pthread_create(&posix_agent->stderr_thread, NULL, ttagent_priv_posix_tail_stderr, agent);
	
			//wait for it to signal that it's started.
			pthread_cond_wait(&posix_agent->tail_signal, &posix_agent->tail_mutex);
		pthread_mutex_unlock(&posix_agent->tail_mutex);
		
	}
}

//------------------------------------------------------------------------------
//		 ttagent_priv_posix_shutdown_errortail
//------------------------------------------------------------------------------

static void 
ttagent_priv_posix_shutdown_errortail(ttagent_context *agent)
{
	ttagent_posix_context *posix_agent;
	
	if (posix_agent = agent->_extra)
	{
		pthread_mutex_lock(&posix_agent->tail_mutex);
			posix_agent->tail_halt = true;
		pthread_mutex_unlock(&posix_agent->tail_mutex);
		pthread_cond_signal(&posix_agent->tail_signal);
		
		pthread_join(posix_agent->stderr_thread, NULL);
		
		pthread_cond_destroy(&posix_agent->tail_signal);
		pthread_mutex_destroy(&posix_agent->tail_mutex);
		free(agent->_extra);
	}

}


//------------------------------------------------------------------------------
//		 ttagent_priv_posix_tail_stderr
//------------------------------------------------------------------------------
//	Didn't think too hard about thread safety here. It assumes, for example,
//	that the file descriptor won't be modified.

static void*
ttagent_priv_posix_tail_stderr(void *agent_param)
{
	ttagent_context			*agent = (ttagent_context*)agent_param;
	ttagent_posix_context	*posix_agent = (ttagent_posix_context*)agent->_extra;
	char					buffer[512] = "";
	short					pid;
	int						bytes_read;
	bool					keep_going;
	char					*search;
	struct timespec			sleep_time;
	
	//sync start.
	pthread_mutex_lock(&posix_agent->tail_mutex);
		pthread_cond_signal(&posix_agent->tail_signal);
		pid = agent->pid;
	pthread_mutex_unlock(&posix_agent->tail_mutex);
	
	// parent thread is running now.
	
	pthread_mutex_lock(&posix_agent->tail_mutex);
	
	//uh, hmm.
	
		fcntl(agent->error_fd,  F_SETFL, O_NONBLOCK);
		keep_going = true;
		do
		{
			memset(buffer, 0, 512);
			bytes_read = read(agent->error_fd, buffer, 511);
			
			if (bytes_read > 0)	//-1 means that there was no data (non-blocking IO).
			{
				if (search = strstr(buffer, "\n"))
				{
					//log formatted newlines
					char		temp[512], *last_search;
					
					last_search = buffer;
					do
					{
						memcpy(temp, last_search, search - last_search);
						temp[search - last_search] = 0;
						
						ttagent_logging_debug(-1, "pid %ld stderr: %s", pid, temp);
						last_search = search;
					}
					while (search = strstr(last_search+1, "\n"));
				}
				else
				{
					//spit it out
					ttagent_logging_debug(-1, "pid %ld stderr: %s", pid, buffer);
				}
				
			}
			else if (bytes_read != 0)
			{
				//sleep a bit.

				gettimeofday((struct timeval*)&sleep_time, NULL);
				//fix nanoseconds. they're very, very important.
				TIMEVAL_TO_TIMESPEC((struct timeval*)&sleep_time, &sleep_time);

				sleep_time.tv_sec += 10;
				pthread_cond_timedwait(&posix_agent->tail_signal, &posix_agent->tail_mutex, &sleep_time);
			}

			pthread_mutex_unlock(&posix_agent->tail_mutex);
				keep_going = !posix_agent->tail_halt;
			pthread_mutex_lock(&posix_agent->tail_mutex);
			
		} while (bytes_read != 0 && keep_going);
		
	pthread_mutex_unlock(&posix_agent->tail_mutex);
	
	return 0;
}
